@@ -1,24 +1,37 @@
 #!/usr/bin/perl
 
 package KiokuDB::Backend::DBI;
+BEGIN {
+  $KiokuDB::Backend::DBI::AUTHORITY = 'cpan:NUFFIN';
+}
+BEGIN {
+  $KiokuDB::Backend::DBI::VERSION = '1.11';
+}
 use Moose;
 
-use MooseX::Types -declare => [qw(ValidColumnName)];
+use Moose::Util::TypeConstraints;
 
-use MooseX::Types::Moose qw(ArrayRef HashRef Str);
+use MooseX::Types -declare => [qw(ValidColumnName SchemaProto)];
+
+use MooseX::Types::Moose qw(ArrayRef HashRef Str Defined);
 
 use Moose::Util::TypeConstraints qw(enum);
 
+use Try::Tiny;
 use Data::Stream::Bulk::DBI;
+use SQL::Abstract;
+use JSON;
+use Scalar::Util qw(weaken refaddr);
 
 use KiokuDB::Backend::DBI::Schema;
-
-use SQL::Abstract;
+use KiokuDB::TypeMap;
+use KiokuDB::TypeMap::Entry::DBIC::Row;
+use KiokuDB::TypeMap::Entry::DBIC::ResultSourceHandle;
+use KiokuDB::TypeMap::Entry::DBIC::ResultSet;
+use KiokuDB::TypeMap::Entry::DBIC::Schema;
 
 use namespace::clean -except => 'meta';
 
-our $VERSION = "1.10";
-
 with qw(
     KiokuDB::Backend
     KiokuDB::Backend::Serialize::Delegate
@@ -28,6 +41,7 @@ with qw(
     KiokuDB::Backend::Role::Query::Simple
     KiokuDB::Backend::Role::Query::GIN
     KiokuDB::Backend::Role::Concurrency::POSIX
+    KiokuDB::Backend::Role::GC
     Search::GIN::Extract::Delegate
 );
 # KiokuDB::Backend::Role::TXN::Nested is not supported by many DBs
@@ -38,9 +52,14 @@ my @reserved_cols = ( @std_cols, 'data' );
 my %reserved_cols = ( map { $_ => 1 } @reserved_cols );
 
 subtype ValidColumnName, as Str, where { not exists $reserved_cols{$_} };
+subtype SchemaProto, as Defined, where {
+    Class::MOP::load_class($_) unless ref;
+    !ref($_) || blessed($_) and $_->isa("DBIx::Class::Schema::KiokuDB");
+};
 
 sub new_from_dsn {
     my ( $self, $dsn, @args ) = @_;
+    @args = %{ $args[0] } if @args == 1 and ref $args[0] eq 'HASH';
     $self->new( dsn => "dbi:$dsn", @args );
 }
 
@@ -56,6 +75,12 @@ sub BUILD {
 
 has '+serializer' => ( default => "json" ); # to make dumps readable
 
+has json => (
+    isa => "Object",
+    is  => "ro",
+    default => sub { JSON->new },
+);
+
 has create => (
     isa => "Bool",
     is  => "ro",
@@ -67,7 +92,6 @@ has 'dsn' => (
     is  => "ro",
 );
 
-
 has [qw(user password)] => (
     isa => "Str",
     is  => "ro",
@@ -154,7 +178,15 @@ has schema => (
     isa => "DBIx::Class::Schema",
     is  => "ro",
     lazy_build => 1,
-    handles => [qw(deploy)],
+    init_arg => "actual_schema",
+    handles  => [qw(deploy kiokudb_entries_source_name)],
+);
+
+has schema_proto => (
+    isa => SchemaProto,
+    is  => "ro",
+    init_arg => "schema",
+    default  => "KiokuDB::Backend::DBI::Schema",
 );
 
 has schema_hook => (
@@ -166,9 +198,11 @@ has schema_hook => (
 sub _build_schema {
     my $self = shift;
 
-    my $schema = KiokuDB::Backend::DBI::Schema->clone;
+    my $schema = $self->schema_proto->clone;
 
-    $schema->source("entries")->add_columns(@{ $self->columns });
+    unless ( $schema->kiokudb_entries_source_name ) {
+        $schema->define_kiokudb_schema( extra_entries_columns => $self->columns );
+    }
 
     if ( $self->has_schema_hook ) {
         my $h = $self->schema_hook;
@@ -202,7 +236,7 @@ has _columns => (
 sub _build__columns {
     my $self = shift;
 
-    my $rs = $self->schema->source("entries");
+    my $rs = $self->schema->source( $self->kiokudb_entries_source_name );
 
     my @user_cols = grep { not exists $reserved_cols{$_} } $rs->columns;
 
@@ -249,9 +283,53 @@ sub _build_sql_abstract {
     SQL::Abstract->new;
 }
 
+# use a Maybe so we can force undef in the builder
+has batch_size => (
+    isa => "Maybe[Int]",
+    is  => "ro",
+    lazy => 1,
+    builder => '_build_batch_size',
+);
+
+sub _build_batch_size {
+    my $self = shift;
+
+    if ($self->storage->sqlt_type eq 'SQLite') {
+        return 999;
+    } else {
+        return undef;
+    }
+}
+
+sub has_batch_size { defined shift->batch_size }
+
+sub register_handle {
+    my ( $self, $kiokudb ) = @_;
+
+    $self->schema->_kiokudb_handle($kiokudb);
+}
+
+sub default_typemap {
+    KiokuDB::TypeMap->new(
+        isa_entries => {
+            # redirect to schema row
+            'DBIx::Class::Row'                => KiokuDB::TypeMap::Entry::DBIC::Row->new,
+
+            # actual serialization
+            'DBIx::Class::ResultSet'          => KiokuDB::TypeMap::Entry::DBIC::ResultSet->new,
+
+            # fake, the entries never get written to the db
+            'DBIx::Class::ResultSourceHandle' => KiokuDB::TypeMap::Entry::DBIC::ResultSourceHandle->new,
+            'DBIx::Class::Schema'             => KiokuDB::TypeMap::Entry::DBIC::Schema->new,
+        },
+    );
+}
+
 sub insert {
     my ( $self, @entries ) = @_;
 
+    return unless @entries;
+
     $self->insert_rows( $self->entries_to_rows(@entries) );
 
     # hopefully we're in a transaction, otherwise this totally sucks
@@ -276,7 +354,7 @@ sub insert {
 sub entries_to_rows {
     my ( $self, @entries ) = @_;
 
-    my ( %insert, %update );
+    my ( %insert, %update, @dbic );
 
     foreach my $t ( \%insert, \%update ) {
         foreach my $col ( @{ $self->_ordered_columns } ) {
@@ -285,12 +363,20 @@ sub entries_to_rows {
     }
 
     foreach my $entry ( @entries ) {
-        my $targ = $entry->prev ? \%update : \%insert;
+        my $id = $entry->id;
 
-        my $row = $self->entry_to_row($entry, $targ);
+        if ( $id =~ /^dbic:schema/ ) {
+            next;
+        } elsif ( $id =~ /^dbic:row:/ ) {
+            push @dbic, $entry->data;
+        } else {
+            my $targ = $entry->prev ? \%update : \%insert;
+
+            my $row = $self->entry_to_row($entry, $targ);
+        }
     }
 
-    return \( %insert, %update );
+    return \( %insert, %update, @dbic );
 }
 
 sub entry_to_row {
@@ -325,18 +411,24 @@ sub entry_to_row {
 }
 
 sub insert_rows {
-    my ( $self, $insert, $update ) = @_;
+    my ( $self, $insert, $update, $dbic ) = @_;
 
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
         if ( $self->extract ) {
             if ( my @ids = map { @{ $_->{id} || [] } } $insert, $update ) {
-                my $del_gin_sth = $dbh->prepare_cached("DELETE FROM gin_index WHERE id IN (" . join(", ", ('?') x @ids) . ")");
 
-                $del_gin_sth->execute(@ids);
+                my $batch_size = $self->batch_size || scalar(@ids);
 
-                $del_gin_sth->finish;
+                my @ids_copy = @ids;
+                while ( my @batch_ids = splice @ids_copy, 0, $batch_size ) {
+                    my $del_gin_sth = $dbh->prepare_cached("DELETE FROM gin_index WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+
+                    $del_gin_sth->execute(@batch_ids);
+
+                    $del_gin_sth->finish;
+                }
             }
         }
 
@@ -367,6 +459,8 @@ sub insert_rows {
 
             $sth->finish;
         }
+
+        $_->insert_or_update for @$dbic;
     });
 }
 
@@ -410,82 +504,256 @@ sub update_index {
     });
 }
 
+sub _parse_dbic_key {
+    my ( $self, $key ) = @_;
+
+    @{ $self->json->decode(substr($key,length('dbic:row:'))) };
+}
+
+sub _part_rows_and_ids {
+    my ( $self, $rows_and_ids ) = @_;
+
+    my ( @rows, @ids, @special );
+
+    for ( @$rows_and_ids ) {
+        if ( /^dbic:schema/ ) {
+            push @special, $_;
+        } elsif ( /^dbic:row:/ ) {
+            push @rows, $_;
+        } else {
+            push @ids, $_;
+        }
+    }
+
+    return \( @rows, @ids, @special );
+}
+
+sub _group_dbic_keys {
+    my ( $self, $keys, $mkey_handler ) = @_;
+
+    my ( %keys, %ids );
+
+    foreach my $id ( @$keys ) {
+        my ( $rs_name, @key ) = $self->_parse_dbic_key($id);
+
+        if ( @key > 1 ) {
+            $mkey_handler->($id, $rs_name, @key);
+        } else {
+            # for other objects we queue up IDs for a single SELECT
+            push @{ $keys{$rs_name} ||= [] }, $key[0];
+            push @{ $ids{$rs_name}  ||= [] }, $id;
+        }
+    }
+
+    return \( %keys, %ids );
+}
+
 sub get {
-    my ( $self, @ids ) = @_;
+    my ( $self, @rows_and_ids ) = @_;
+
+    return unless @rows_and_ids;
 
     my %entries;
 
-    $self->dbh_do(sub {
-        my ( $storage, $dbh ) = @_;
+    my ( $rows, $ids, $special ) = $self->_part_rows_and_ids(\@rows_and_ids);
 
-        my $sth;
+    if ( @$ids ) {
+        $self->dbh_do(sub {
+            my ( $storage, $dbh ) = @_;
 
-        if ( @ids ) {
-            $sth = $dbh->prepare_cached("SELECT id, data FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
-            $sth->execute(@ids);
-        } else {
-            $sth = $dbh->prepare_cached("SELECT id, data FROM entries");
-            $sth->execute;
+            my @ids_copy = @$ids;
+
+            my $batch_size = $self->batch_size || scalar(@$ids);
+
+            while ( my @batch_ids = splice(@ids_copy, 0, $batch_size) ) {
+                my $sth = $dbh->prepare_cached("SELECT id, data FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+                $sth->execute(@batch_ids);
+
+                $sth->bind_columns( \my ( $id, $data ) );
+
+                # not actually necessary but i'm keeping it around for reference:
+                #my ( $id, $data );
+                #use DBD::Pg qw(PG_BYTEA);
+                #$sth->bind_col(1, \$id);
+                #$sth->bind_col(2, \$data, { pg_type => PG_BYTEA });
+
+                while ( $sth->fetch ) {
+                    $entries{$id} = $data;
+                }
+            }
+        });
+    }
+
+    if ( @$rows ) {
+        my $schema = $self->schema;
+
+        my $err = \"foo";
+        my ( $rs_keys, $rs_ids ) = try {
+            $self->_group_dbic_keys( $rows, sub {
+                my ( $id, $rs_name, @key ) = @_;
+
+                # multi column primary keys need 'find'
+                my $obj = $schema->resultset($rs_name)->find(@key) or die $err; # die to stop search
+
+                $entries{$id} = KiokuDB::Entry->new(
+                    id    => $id,
+                    class => ref($obj),
+                    data  => $obj,
+                );
+            });
+        } catch {
+            die $_ if ref $_ and refaddr($_) == refaddr($err);
+        } or return;
+
+        foreach my $rs_name ( keys %$rs_keys ) {
+            my $rs = $schema->resultset($rs_name);
+
+            my $ids = $rs_ids->{$rs_name};
+
+            my @objs;
+
+            if ( @$ids == 1 ) {
+                my $id = $ids->[0];
+
+                my $obj = $rs->find($rs_keys->{$rs_name}[0]) or return;
+
+                $entries{$id} = KiokuDB::Entry->new(
+                    id => $id,
+                    class => ref($obj),
+                    data => $obj,
+                );
+            } else {
+                my ($pk) = $rs->result_source->primary_columns;
+
+                my $keys = $rs_keys->{$rs_name};
+
+                my @objs = $rs->search({ $pk => $keys })->all;
+
+                return if @objs != @$ids;
+
+                # this key lookup is because it's not returned in the same order
+                my %pk_to_id;
+                @pk_to_id{@$keys} = @$ids;
+
+                foreach my $obj ( @objs ) {
+                    my $id = $pk_to_id{$obj->id};
+                    $entries{$id} = KiokuDB::Entry->new(
+                        id    => $id,
+                        class => ref($obj),
+                        data  => $obj,
+                    );
+                }
+            }
         }
+    }
 
-        $sth->bind_columns( \my ( $id, $data ) );
+    for ( @$special ) {
+        $entries{$_} = KiokuDB::Entry->new(
+            id => $_,
+            $_ eq 'dbic:schema'
+                ? ( data => $self->schema,
+                    class => "DBIx::Class::Schema" )
+                : ( data => undef,
+                    class => "DBIx::Class::ResultSourceHandle" )
+        );
+    }
 
-        # not actually necessary but i'm keeping it around for reference:
-        #my ( $id, $data );
-        #use DBD::Pg qw(PG_BYTEA);
-        #$sth->bind_col(1, \$id);
-        #$sth->bind_col(2, \$data, { pg_type => PG_BYTEA });
+    return if @rows_and_ids != keys %entries; # ->rows only works after we're done
 
-        while ( $sth->fetch ) {
-            $entries{$id} = $data;
-        }
-    });
-
-    return if @ids != keys %entries; # ->rows only works after we're done
-
-    return map { $self->deserialize($_) } @entries{@ids};
+    map { ref($_) ? $_ : $self->deserialize($_) } @entries{@rows_and_ids};
 }
 
 sub delete {
     my ( $self, @ids_or_entries ) = @_;
 
+    # FIXME special DBIC rows
+
+    my @ids = map { ref($_) ? $_->id : $_ } @ids_or_entries;
+
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
-        my @ids = map { ref($_) ? $_->id : $_ } @ids_or_entries;
+        my $batch_size = $self->batch_size || scalar(@ids);
 
-        if ( $self->extract ) {
-            # FIXME rely on cascade delete?
-            my $sth = $dbh->prepare_cached("DELETE FROM gin_index WHERE id IN (" . join(", ", ('?') x @ids) . ")");
-            $sth->execute(@ids);
+        my @ids_copy = @ids;
+        while ( my @batch_ids = splice @ids_copy, 0, $batch_size ) {
+            if ( $self->extract ) {
+                # FIXME rely on cascade delete?
+                my $sth = $dbh->prepare_cached("DELETE FROM gin_index WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+                $sth->execute(@batch_ids);
+                $sth->finish;
+            }
+
+            my $sth = $dbh->prepare_cached("DELETE FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+            $sth->execute(@batch_ids);
             $sth->finish;
         }
-
-        my $sth = $dbh->prepare_cached("DELETE FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
-        $sth->execute(@ids);
-        $sth->finish;
     });
 
     return;
 }
 
 sub exists {
-    my ( $self, @ids ) = @_;
+    my ( $self, @rows_and_ids ) = @_;
+
+    return unless @rows_and_ids;
+
+    my $schema = $self->schema;
 
     my %entries;
 
-    $self->dbh_do(sub {
-        my ( $storage, $dbh ) = @_;
+    my ( $rows, $ids, $special ) = $self->_part_rows_and_ids(\@rows_and_ids);
 
-        my $sth = $dbh->prepare_cached("SELECT id FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
-        $sth->execute(@ids);
+    if ( @$ids ) {
+        $self->dbh_do(sub {
+            my ( $storage, $dbh ) = @_;
 
-        $sth->bind_columns( \( my $id ) );
+            my $batch_size = $self->batch_size || scalar(@$ids);
 
-        $entries{$id} = undef while $sth->fetch;
-    });
+            my @ids_copy = @$ids;
+            while ( my @batch_ids = splice @ids_copy, 0, $batch_size ) {
+                my $sth = $dbh->prepare_cached("SELECT id FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+                $sth->execute(@batch_ids);
 
-    map { exists $entries{$_} } @ids;
+                $sth->bind_columns( \( my $id ) );
+
+                $entries{$id} = 1 while $sth->fetch;
+            }
+        });
+    }
+
+    if ( @$rows ) {
+        my ( $rs_keys, $rs_ids ) = $self->_group_dbic_keys( $rows, sub {
+            my ( $id, $rs_name, @key ) = @_;
+            $entries{$id} = defined $schema->resultset($rs_name)->find(@key); # FIXME slow
+        });
+
+        foreach my $rs_name ( keys %$rs_keys ) {
+            my $rs = $schema->resultset($rs_name);
+
+            my $ids = $rs_ids->{$rs_name};
+            my $keys = $rs_keys->{$rs_name};
+
+            my ( $pk ) = $rs->result_source->primary_columns;
+
+            my @exists = $rs->search({ $pk => $keys })->get_column($pk)->all;
+
+            my %pk_to_id;
+            @pk_to_id{@$keys} = @$ids;
+
+            @entries{@pk_to_id{@exists}} = ( (1) x @exists );
+        }
+    }
+
+    for ( @$special ) {
+        if ( $_ eq 'dbic:schema' ) {
+            $entries{$_} = 1;
+        } elsif ( /^dbic:schema:(.*)/ ) {
+            $entries{$_} = defined try { $schema->source($1) };
+        }
+    }
+
+    return @entries{@rows_and_ids};
 }
 
 sub txn_begin    { shift->storage->txn_begin(@_) }
@@ -614,18 +882,28 @@ sub insert_entry {
     die "Insertion to the GIN index is handled implicitly";
 }
 
-sub tables_exist {
-    my $self = shift;
+sub _table_info {
+    my ( $self, $catalog, $schema, $table ) = @_;
 
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
         my $filter = ( $self->storage->sqlt_type eq 'SQLite' ? '%' : '' );
 
-        return ( @{ $dbh->table_info($filter, $filter, 'entries', 'TABLE')->fetchall_arrayref } > 0 );
+        foreach my $arg ( $catalog, $schema, $table ) {
+            $arg = $filter unless defined $arg;
+        }
+
+        $dbh->table_info($catalog, $schema, $table, 'TABLE')->fetchall_arrayref;
     });
 }
 
+sub tables_exist {
+    my $self = shift;
+
+    return ( @{ $self->_table_info(undef, undef, 'entries') } > 0 );
+}
+
 sub create_tables {
     my $self = shift;
 
@@ -657,6 +935,25 @@ sub DEMOLISH {
     }
 }
 
+sub new_garbage_collector {
+    my ( $self, %args ) = @_;
+
+    if ( grep { $_ !~ /^(?:entries|gin_index)/ } map { $_->[2] } @{ $self->_table_info } ) {
+        die "\nRefusing to GC a database with additional tables.\n\nThis is ecause the root set and referencing scheme might be ambiguous (it's not yet clear what garbage collection should actually do on a mixed schema).\n";
+    } else {
+        my $cmd = $args{command};
+        my $class = $args{class} || $cmd ? $cmd->class : "KiokuDB::GC::Naive";
+
+        Class::MOP::load_class($class);
+
+        return $class->new(
+            %args,
+            backend => $self,
+            ( $cmd ? ( verbose => $cmd->verbose ) : $cmd ),
+        );
+    }
+}
+
 __PACKAGE__->meta->make_immutable;
 
 __PACKAGE__
@@ -718,6 +1015,16 @@ will be C<NULL>.
 These columns are only used for lookup purposes, only C<data> is consulted when
 loading entries.
 
+=head1 DBIC INTEGRATION
+
+This backend is layered on top of L<DBIx::Class::Storage::DBI> and reused
+L<DBIx::Class::Schema> for DDL.
+
+Because of this objects from a L<DBIx::Class::Schema> can refer to objects in
+the KiokuDB entries table, and vice versa.
+
+For more details see L<DBIx::Class::Schema::KiokuDB>.
+
 =head1 SUPPORTED DATABASES
 
 This driver has been tested with MySQL 5 (4.1 should be the minimal supported
@@ -873,6 +1180,16 @@ Defaults to
 
     { on_connect_call => $self->on_connect_call }
 
+=item batch_size
+
+SQL that deals with entries run in batches of the amount provided in
+C<batch_size>. If it is not provided, the statements will run in a single
+batch.
+
+This solves the issue with SQLite where lists can only handle 999
+elements at a time. C<batch_size> will be set to 999 by default if the
+driver in use is SQLite.
+
 =back
 
 =head1 METHODS
@@ -898,6 +1215,23 @@ Drops the C<entries> and C<gin_index> tables.
 
 =back
 
+=head1 TROUBLESHOOTING
+
+=head2 I get C<unexpected end of string while parsing JSON string>
+
+You are problably using MySQL, which comes with a helpful data compression
+feature: when your serialized objects are larger than the maximum size of a
+C<BLOB> column MySQL will simply shorten it for you.
+
+Why C<BLOB> defaults to 64k, and how on earth someone would consider silent
+data truncation a sane default I could never fathom, but nevertheless MySQL
+does allow you to disable this by setting the "strict" SQL mode in the
+configuration.
+
+To resolve the actual problem (though this obviously won't repair your lost
+data), alter the entries table so that the C<data> column uses the nonstandard
+C<LONGBLOB> datatype.
+
 =head1 VERSION CONTROL
 
 L<http://github.com/nothingmuch/kiokudb-backend-dbi>
